import javax.vecmath.*;

//
// a class with static functions to compute various things for
// camera projection matrices P 
//
class CameraTools {
    // M -> QR (M is square, Q is orthogonal, R is upper triangular)
    static void qr(GMatrix M, GMatrix Q, GMatrix R) {
        GMatrix V = new GMatrix(M);
        GVector vi = new GVector(V.getNumCol());
        GVector qi = new GVector(Q.getNumCol());

        Q.setZero();
        R.setZero();

        for (int i = 0; i < M.getNumCol(); i++) {
            V.getColumn(i, vi);
            double rii = vi.norm();  R.setElement(i,i,rii);
            qi.scale(1.0/rii, vi);   Q.setColumn(i, qi);

            for (int j = i+1; j < M.getNumCol(); j++) {
                V.getColumn(j, vi);
                double rij = qi.dot(vi);    R.setElement(i,j,rij);
                vi.scaleAdd(-rij, qi, vi);  V.setColumn(j, vi);
            }
        }
    }

    // M -> RQ (M is square, R is upper triangular, Q is orthogonal)
    static void rq(GMatrix M, GMatrix R, GMatrix Q) {
        GMatrix V = new GMatrix(M);
        GVector vi = new GVector(V.getNumRow());
        GVector qi = new GVector(Q.getNumRow());

        R.setZero();
        Q.setZero();

        for (int i = M.getNumRow()-1; i >=0; i--) {
            V.getRow(i, vi);
            double rii = vi.norm();  R.setElement(i,i,rii);
            qi.scale(1.0/rii, vi);   Q.setRow(i, qi);

            for (int j = i-1; j >= 0; j--) {
                V.getRow(j, vi);
                double rij = qi.dot(vi);    R.setElement(j,i,rij);
                vi.scaleAdd(-rij, qi, vi);  V.setRow(j, vi);
            }
        }
    }
    
    // returns a 3-vector of the camera center in projection matrix P
    public static GVector getCenter(GMatrix P) {
        GMatrix Minv = new GMatrix(P); Minv.setSize(3,3); Minv.invert(); // M^-1
        GVector p3 = new GVector(3); P.getColumn(3, p3);
        GVector C = new GVector(3); C.mul(Minv, p3); C.negate(); // camera center
        return C;
    }

    // returns a 3x3 matrix of the rotation in projection matrix P
    public static GMatrix getRotationMatrix(GMatrix P) {
        GMatrix M = new GMatrix(P); M.setSize(3,3);
        GMatrix R = new GMatrix(3,3);
        GMatrix Q = new GMatrix(3,3);
        rq(M, R, Q);
        return Q;
    }

    // returns a the 3-vector rotation axis, whose length = angle in rad, found in projection matrix P
    public static GVector getRotationAxis(GMatrix P) {
        GMatrix R = getRotationMatrix(P);
        GVector a = new GVector(3);
        
        // axis is eigenvector with eigenvalue = 1
        // find it with inverse iteration method
        GMatrix RI = new GMatrix(R);
        RI.identityMinus();
        RI.negate();
        RI.invert();
        a.set(new double[] {1,0,0});
        for (int i = 0; i < 30 /*how many?*/; i++) {
            a.mul(RI,a);
            a.normalize();
        }

        // find angle with R, a and v...
        double x = R.trace() - 1;
        GVector v = new GVector(new double[] 
                {R.getElement(2,1) - R.getElement(1,2),
                 R.getElement(0,2) - R.getElement(2,0),
                 R.getElement(1,0) - R.getElement(0,1)},3);
        double y = a.dot(v);
        double angle = Math.atan2(y,x);

        a.scale(angle);
        return a;
    }

    // do spherical linear interpolation, v1 and v2 need to be unit vectors, and t in [0,1]
    public static GVector slerp(GVector v1, GVector v2, double t) {
        double v1dotv2 = v1.dot(v2);
	GVector v3 = new GVector(v1);
	v3.scale(v1dotv2);
	v3.sub(v2, v3);
	v3.normalize();

	double theta = t*Math.acos(v1dotv2);
	v3.scale(Math.sin(theta));
	v3.scaleAdd(Math.cos(theta), v1, v3);

	return v3;
    }

    // returns a 3x3 rotation matrix corresponding to the 3-vector rotation axis, whose length = angle in rad
    public static GMatrix makeRotationMatrix(GVector axis) {
        GVector a = new GVector(axis);
        double angle = a.norm();
        a.normalize();

        GMatrix symm = new GMatrix(3,3);
        symm.mul(a, a);
        
        double ax = a.getElement(0);
        double ay = a.getElement(1);
        double az = a.getElement(2);
        GMatrix skew = new GMatrix(3,3, new double[] 
                {  0, -az,  ay, 
                  az,   0, -ax, 
                 -ay,  ax,   0});

        // rotation matrix = aa^T * (1-cos(angle)) + a_skew * sin(angle) + I * cos(angle)
        GMatrix scale = new GMatrix(3,3);
        scale.setScale(1-Math.cos(angle));
        symm.mul(scale);

        scale.setScale(Math.sin(angle));
        skew.mul(scale);

        scale.setScale(Math.cos(angle));

        symm.add(skew);
        symm.add(scale);

        return symm;
    }

}
