import java.awt.*;
import java.util.*;
import java.io.*;
import java.awt.image.*;
import javax.imageio.*;
import javax.vecmath.*;

class SteadiMonkey {
    // very good random number generator
    static final java.security.SecureRandom random = new java.security.SecureRandom();

    public static final double ZMIN = 20;
    public static final double ZMAX = 70;
    public static final double DELTAZ = 0.1;

    //
    // computes E_photo with color set, and for the current virtual color estimate V
    //
    public static double Ephoto(int colors[][], ColorRGB V) {
        double Emin = Double.MAX_VALUE;
        for (int z = 0; z < colors.length; z++) {
            if (colors[z] == null) 
                continue; // no color info for this depth

            double E = 0;
            for (int i = 0; i < colors[z].length; i++) {
                E += V.distance(colors[z][i]);
            }
            E /= colors[z].length;
	    Emin = Math.min(Emin, E);
        }
        return Emin;
    }

    //
    // computes the photoconsitency modes using images and projections and projection matrix P of the view of interest
    // ... very very expensive ... so writes down in a file the modes for every line
    //
    public static Vector<ColorRGB>[][] computeModes(Vector<BufferedImage> images, Vector<GMatrix> projections, GMatrix P, Rectangle rect) throws IOException {
        GMatrix Minv = new GMatrix(P); Minv.setSize(3,3); Minv.invert(); // M^-1
        GVector C = CameraTools.getCenter(P); C.setSize(4); C.setElement(3,1);       // homogenized camera center

        int width = images.get(0).getWidth();
        int height = images.get(0).getHeight();
        if (rect == null) {
            rect = new Rectangle(0,0,width,height);
        }

        GVector D = new GVector(3); // direction vector
        GVector X = new GVector(4); // finite 3D point
        GVector xi = new GVector(3); // projection into 2D image plane
        final int colors[][] = new int[(int)((ZMAX-ZMIN)/DELTAZ)][images.size()]; // color set

        // stick that into a method with @SuppressWarnings("unchecked")
        Vector<ColorRGB> modes[][] = (Vector<ColorRGB>[][]) new Vector[height][width];

        for (int y = rect.y; y < rect.y+rect.height; y++) {
            for (int x = rect.x; x < rect.x+rect.width; x++) {
                D.setSize(3); D.set(new double[] {x,y,1}); D.mul(Minv, D);
                D.normalize(); D.setSize(4); D.setElement(3,0); // normalize and homogenize it

                for (double z = ZMIN; z < ZMAX; z+=DELTAZ) {
                    X.set(D); X.scale(z); X.add(C); // new 3D point

                    int k = 0;
                    int zidx = (int)((z-ZMIN)/DELTAZ);
                    colors[zidx] = new int[images.size()];
                    for (int i = 0; i < images.size(); i++) {
                        xi.mul(projections.get(i), X); // compute projection of X into image plane i
                        int xix = (int)Math.round(xi.getElement(0)/xi.getElement(2));
                        int xiy = (int)Math.round(xi.getElement(1)/xi.getElement(2));

                        // add color to color set if the xi ends up in the image plane
                        if (xix >= 0 && xix < width && xiy >= 0 && xiy < height) {
                            colors[zidx][k++] = images.get(i).getRGB(xix, xiy);
                        }
                    }
                    // resize down array of colors for this depth
                    if (k == 0) {
                        colors[zidx] = null;
                    } else {
                        int depthArray[] = new int[k];
                        System.arraycopy(colors[zidx], 0, depthArray, 0, k);
                        colors[zidx] = depthArray;
                    }
                }
                
                double maxDistance = 5;
                int randomRestarts = 20;
                double initialAlphaIncrement = 15;
                ColorRGB V = new ColorRGB(0,0,0);
                ColorRGB Vp = new ColorRGB(0,0,0);
                ColorRGB Vm = new ColorRGB(0,0,0);
                modes[y][x] = new Vector<ColorRGB>(randomRestarts);
                // do 20 random restarts of gradient descent on E_photo
                for (int i = 0; i < randomRestarts; i++) {
                    // works better if we start with random gray values than color values
                    //V.set(random.nextDouble()*255, random.nextDouble()*255, random.nextDouble()*255);
                    double gray = random.nextDouble()*255;
                    V.set(gray, gray, gray);

                    // perform steepest descent on E_photo with initial virtual color estimate V
                    while (true) {
                        double E = Ephoto(colors, V);

                        Vp.set(V); Vp.setComponent(0, Vp.getComponent(0)+1);
                        Vm.set(V); Vm.setComponent(0, Vm.getComponent(0)-1);
                        double grad0 = Ephoto(colors, Vp) - Ephoto(colors, Vm);

                        Vp.set(V); Vp.setComponent(1, Vp.getComponent(1)+1);
                        Vm.set(V); Vm.setComponent(1, Vm.getComponent(1)-1);
                        double grad1 = Ephoto(colors, Vp) - Ephoto(colors, Vm);

                        Vp.set(V); Vp.setComponent(2, Vp.getComponent(2)+1);
                        Vm.set(V); Vm.setComponent(2, Vm.getComponent(2)-1);
                        double grad2 = Ephoto(colors, Vp) - Ephoto(colors, Vm);

                        double increment = initialAlphaIncrement;
                        double alpha = increment;
                        while (true) {
                            Vp.setComponent(0, V.getComponent(0) - alpha*grad0);
                            Vp.setComponent(1, V.getComponent(1) - alpha*grad1);
                            Vp.setComponent(2, V.getComponent(2) - alpha*grad2);
                            double newE = Ephoto(colors, Vp);

                            if (newE < E) {
                                if ((newE - E)/E > -0.001)
                                    break;

                                // we are going down, good
                                E = newE;
                                alpha += increment;
                            } else {
                                if ((newE - E)/E < 0.001) {
                                    alpha -= increment;
                                    Vp.setComponent(0, V.getComponent(0) - alpha*grad0);
                                    Vp.setComponent(1, V.getComponent(1) - alpha*grad1);
                                    Vp.setComponent(2, V.getComponent(2) - alpha*grad2);
                                    break;
                                }

                                // oops, we are going up, backtrack
                                // and reduce increment
                                alpha -= increment;
                                increment /= 2;
                                alpha += increment;
                            }
                        }

                        if (V.distance(Vp) < 0.05)  {
                            // distance < 0.05 -> mostly converged, break
                            break;
                        }
                        V.set(Vp);
                    }

                    // cluster minima for efficiency
                    boolean foundCluster = false;
                    for (int k = 0; k < modes[y][x].size(); k++) {
                        ColorRGB modek = modes[y][x].get(k);
                        if (modek.distance(V) < maxDistance) {
                            modek.update(V, 0.1); // update modek's mean, alpha = 0.1?
                            foundCluster = true;
                            break;
                        }
                    }
                    if (!foundCluster) {
                        modes[y][x].add(new ColorRGB(V));
                    }
                }
                // sort modes starting with the one with least energy
                // for use in runPriorIteration
                Collections.sort(modes[y][x], new Comparator<ColorRGB>() {
                    public int compare(ColorRGB c1, ColorRGB c2) {
                        double E1 = Ephoto(colors, c1);
                        double E2 = Ephoto(colors, c2);
                        return E1 < E2 ? -1 : 1; // lower energy is better, thus first -> -1
                    }
                });

                System.out.println("(" + x + "," + y + ") " + modes[y][x].size() + ": " + modes[y][x].get(0));
            }
            
            // store each line on disk for later reuse
            FileOutputStream out = new FileOutputStream("modes_" + y + ".object");
            ObjectOutputStream oos = new ObjectOutputStream(out);
            oos.writeObject(modes[y]);
            oos.flush();
        }

        return modes;
    }

    //
    // reads the modes from previously computed  files 
    //
    public static Vector<ColorRGB>[][] readOutModes(Vector<BufferedImage> images, Rectangle rect) throws IOException {
        int height = images.get(0).getHeight();
        int width = images.get(0).getWidth();
        if (rect == null) {
            rect = new Rectangle(0,0,width,height);
        }
        // stick that into a method with @SuppressWarnings("unchecked")
        Vector<ColorRGB> modes[][] = (Vector<ColorRGB>[][]) new Vector[height][width];

        for (int y = rect.y; ; y++) {
            try {
                FileInputStream in = new FileInputStream("modes_" + y + ".object");
                ObjectInputStream ois = new ObjectInputStream(in);
                modes[y] = (Vector<ColorRGB>[]) ois.readObject();
            } catch (FileNotFoundException e) {
                System.err.println("Finished loading modes at y=" + y);
                return modes;
            } catch (ClassNotFoundException e) {
                System.err.println("ClassNotFoundException at y=" + y);
                return modes;
            }
        }
    }

    //
    // returns the mode with least energy from modes into an initial "maximum likelihood" estimate
    //
    public static BufferedImage getMLImage(Rectangle rect, Vector<ColorRGB> modes[][]) {
        int height = modes.length;
        int width = modes[0].length;
        if (rect == null) {
            rect = new Rectangle(0,0,width,height);
        }

        BufferedImage MLImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);

        for (int y = rect.y; y < rect.y+rect.height; y++) {
            for (int x = rect.x; x < rect.x+rect.width; x++) {      
                MLImage.setRGB(x,y,modes[y][x].get(0).getRGB());
            }
        }
        return MLImage;
    }

    //
    // runs one iteration of the texture prior using images, projections, new view matrix P, modes, the current estimage ImageV
    //
    public static void runPriorIteration(Vector<BufferedImage> images, Vector<GMatrix> projections, GMatrix P, Rectangle rect, Vector<ColorRGB> modes[][], BufferedImage ImageV) {
        double lambda = 1.0; // ??

        GMatrix Minv = new GMatrix(P); Minv.setSize(3,3); Minv.invert(); // M^-1
        GVector C = CameraTools.getCenter(P); C.setSize(4); C.setElement(3,1);       // homogenized camera center

        int height = ImageV.getHeight();
        int width = ImageV.getWidth();
        if (rect == null) {
            rect = new Rectangle(0,0,width,height);
        }

        GVector D = new GVector(3); // direction vector
        GVector X = new GVector(4); // finite 3D point
        GVector xi = new GVector(3); // projection into 2D image plane

        for (int y = rect.y; y < rect.y+rect.height; y++) {
            for (int x = rect.x; x < rect.x+rect.width; x++) {
                D.setSize(3); D.set(new double[] {x,y,1}); D.mul(Minv, D);
                D.normalize(); D.setSize(4); D.setElement(3,0); // normalize and homogenize it

                // find patches to use from images and projections
                Patch patches[] = new Patch[(int)((ZMAX-ZMIN)/DELTAZ*images.size())+1];
                int k = 0;
                for (double z = ZMIN; z < ZMAX; z+=DELTAZ) {
                    X.set(D); X.scale(z); X.add(C); // new 3D point

                    for (int i = 0; i < images.size(); i++) {
                        xi.mul(projections.get(i), X);
                        int xix = (int)Math.round(xi.getElement(0)/xi.getElement(2));
                        int xiy = (int)Math.round(xi.getElement(1)/xi.getElement(2));

                        if (xix >= 2 && xix < width-2 && xiy >= 2 && xiy < height-2) {
                            Patch patch = new Patch(images.get(i), xix-2, xiy-2, 5, 5);
                            patches[k++] = patch;
                        }
                    }
                }

                // run through all patches and find best matching one
                Patch window = new Patch(ImageV, x-2, y-2, 5, 5);
                window.set(2,2,0);
                double minDistance = Double.MAX_VALUE;
                int mincolor = -1;
                for (int i = 0; i < k; i++) {
                    int color = patches[i].get(2,2);
                    patches[i].set(2,2,0);
                    double distance = window.distance(patches[i]);
                    if (distance < minDistance) {
                        minDistance = distance;
                        mincolor = color;
                    }
                }

                // update the color of the pixel with the ICM approximation equation
                ColorRGB oldColor = new ColorRGB(ImageV.getRGB(x,y));
                ColorRGB patchColor = new ColorRGB(mincolor);
                ColorRGB newColor = oldColor.plus(patchColor.times(lambda)).times(1.0/(1.0 + lambda));
                ColorRGB minColor = null;
                minDistance = Double.MAX_VALUE;
                for (int i = 0; i < modes[y][x].size(); i++) {
                    ColorRGB color = modes[y][x].get(i);
                    double distance = color.distance(newColor);
                    if (distance < minDistance) {
                        minColor = color;
                        minDistance = distance;
                    }
                }
                ImageV.setRGB(x, y, minColor.getRGB());

                System.out.println(k + " (" + x + " " + y + "): " + oldColor + " -> " + minColor);
            }
        }
    }

    //
    // main method that loads the monkey images from directory "monkey",
    // computes modes, runs the prior, etc etc
    //
    public static void main(String[] args) throws IOException {
        random.setSeed(1);

        Vector<BufferedImage> images = new Vector<BufferedImage>();
        Vector<GMatrix> projections = new Vector<GMatrix>(); 
        GMatrix K = new GMatrix(3,3); K.setZero();  // camera matrix K
        GMatrix M = new GMatrix(3,3);               // temporary 3x3 submatrix of P
        GMatrix Ki = new GMatrix(3,3);              // temporary K from matrix projections[i]
        GMatrix Q = new GMatrix(3,3);               // temporary 3x3 orthogonal matrix
        Scanner s = new Scanner(new BufferedReader(new FileReader(args[1])));
        for (int i = 0; s.hasNext(); i++) {
            // load monkey images
            BufferedImage image = ImageIO.read(new File(String.format("%s/rdc.%03d.png",args[0],i)));
            BufferedImage image2 = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_INT_ARGB);
            image2.createGraphics().drawImage(image, 0, 0, null);
            images.add(image2);

            // load monkey matrices
            double[] matrix = new double[3*4];
            for (int j = 0; j < 3*4; j++)
                matrix[j] = s.nextDouble();
            GMatrix projection = new GMatrix(3,4,matrix);
            projections.add(projection);

            // find camera matrix K
            projection.copySubMatrix(0,0,3,3,0,0,M);
            CameraTools.rq(M, Ki, Q);
            K.add(Ki);
        }
        s.close();

        // find mean camera matrix K
        for (int i = 0; i < 3; i++)
            for (int j = 0; j < 3; j++)
                K.setElement(i,j,K.getElement(i,j)/projections.size());
        System.out.println(K);


        GMatrix P = null; // as projection matrix for the new virtual view...
        if (true) {
            P = projections.get(76);   // ...simply use the matrix from view number 76
            projections.remove(76);    // and remove it from the projections
            images.remove(76);         // and images
            for (int i = 0; i < 62; i++) {
                projections.remove(0); // only keep 26 last projections 
                images.remove(0);      // and images
            }
        } else {
            // ... or we could also try to come up with a new projectrion matrix
            // this is code to interpolate between first and last matrix...
	    // Note: only ~26 closest views should be used, this is not done
	    // at the moment.

	    double t = 0.5; // in the middle


	    //
	    // Angle interpolation... quaternions should be used for efficiency here if required
	    //
            GVector a1 = CameraTools.getRotationAxis(projections.get(0));
            GVector a2 = CameraTools.getRotationAxis(projections.get(projections.size()-1));
            GVector a1unit = new GVector(a1); a1unit.normalize();
            GVector a2unit = new GVector(a2); a2unit.normalize();
            double angle1 = a1.norm();
            double angle2 = a2.norm();
            
            // make sure the axes are on the same side before interpolation
            if (a1unit.dot(a2unit) < 0) { a1unit.negate(); angle1 = -angle1; }
            GVector a3 = CameraTools.slerp(a1unit, a2unit, t);

            // only process positive angles	    
	    if (angle1 < 0) angle1 += 2*Math.PI;
	    if (angle2 < 0) angle2 += 2*Math.PI;
	    
	    // interpolate the "short way" between angles
	    if (Math.abs(angle1-angle2) > Math.PI)
                a3.scale(((angle1+angle2)-Math.PI)*t);
	    else
                a3.scale((angle1+angle2)*t);

            GMatrix R = CameraTools.makeRotationMatrix(a3);
            System.out.println(R);


	    //
	    // Camera center interpolation
	    //
            GVector C1 = CameraTools.getCenter(projections.get(0)); 
            GVector C2 = CameraTools.getCenter(projections.get(projections.size()-1));
            C1.interpolate(C2, t);
            GMatrix T = new GMatrix(3,4); // sub 3x3 matrix is identity
            C1.negate(); T.setColumn(3, C1);
            System.out.println(T);

	    
	    // new camera projection matrix
            P = new GMatrix(3,4); P.mul(R, T); P.mul(K, P); // P = KRT
        }

        System.out.println(P);

        Rectangle rect = new Rectangle(140,125,295,220);
//        Rectangle rect = new Rectangle(270,280,5,10);
//        Rectangle rect = new Rectangle(247,224,8,8);
        Vector<ColorRGB> modes[][] = computeModes(images, projections, P, rect);
//        Vector<ColorRGB> modes[][] = readOutModes(images, rect); // if previously precomputed...
        BufferedImage ImageV = getMLImage(rect, modes);
        ImageIO.write(ImageV, "PNG", new File("MLimage.png"));
        for (int it = 0; it < 10; it++) {
            runPriorIteration(images, projections, P, rect, modes, ImageV);
            ImageIO.write(ImageV, "PNG", new File("MAPimage" + it + ".png"));
        }
    }
}
